#include "stdlib.h"
#include "stdio.h"
#include "time.h"

#include "dd.h"

DirectDrawWin ddwin;

//#define DOT_CNT 5000 * 2
//#define SCREEN_WIDTH  256
//#define SCREEN_HEIGHT 256

#define DOT_CNT 15000 * 2
#define SCREEN_WIDTH  512
#define SCREEN_HEIGHT 512

struct dot {
    u_short x;
    u_short y;
    char dir_n;
    char dir_s;
    char dir_e;
    char dir_w;
};
struct { u_char r, g, b; } colors[7]; // the 7 colors dots can be

dot dots[DOT_CNT]; // these are all the dots we are messing with
dot *positions[SCREEN_WIDTH * SCREEN_HEIGHT]; // this is a screen->dot mapping

int bpp;
int pitch;

inline dot *
get_dot(u_short x, u_short y) {
    return positions[x + y * SCREEN_WIDTH];
}

inline void
set_dot(u_char *base, u_short x, u_short y, dot *d, bool updatepos = true)
{
    if (updatepos) positions[x + y * SCREEN_WIDTH] = d;
    base += (y * pitch) + (x * bpp);
    if (d == NULL) {
	ddwin.set_pixel(base, 0, 0, 0);
    } else {
	int v = d->dir_s - d->dir_n + 3;
	ddwin.set_pixel(base, colors[v].r, colors[v].g, colors[v].b);
    }
}

void
initsand(void)
{
    u_char *ddbuf = ddwin.GetBuffer(&pitch);
    bpp = ddwin.get_bits_per_pixel() / 8;
    memset(ddbuf, 0, pitch * SCREEN_HEIGHT);

    memset(positions, 0, sizeof(positions));

    colors[0].r = 0xff; colors[0].g = 0x00; colors[0].b = 0x00;
    colors[1].r = 0xff; colors[1].g = 0x00; colors[1].b = 0x55;
    colors[2].r = 0xff; colors[2].g = 0x00; colors[2].b = 0xaa;
    colors[3].r = 0xff; colors[3].g = 0x00; colors[3].b = 0xff;
    colors[4].r = 0xaa; colors[4].g = 0x00; colors[4].b = 0xff;
    colors[5].r = 0x55; colors[5].g = 0x00; colors[5].b = 0xff;
    colors[6].r = 0x00; colors[6].g = 0x00; colors[6].b = 0xff;


    //
    // Init the dots. Alternate with a dot going down and a dot going up until
    // all dots are placed. Alternating like this makes movement more fair
    // initially (instead of all down dots moving first, and then up dots).
    //

    dot *d = dots;
    int i = 0;
    while (i < DOT_CNT) {
	d->dir_e = 1;
	d->dir_w = 1;
	if (i % 2) {
	    d->dir_n = 0;
	    d->dir_s = 2;
	} else {
	    d->dir_n = 2;
	    d->dir_s = 0;
	}

	while (1) { // loop until we find an empty pos for the dot
	    d->x = rand() % SCREEN_WIDTH;
	    if (i % 2) {
		d->y = rand() % (SCREEN_HEIGHT / 3);
	    } else {
		d->y = (rand() % (SCREEN_HEIGHT / 3)) + (2 * SCREEN_HEIGHT / 3);
	    }

	    dot *p = get_dot(d->x, d->y);
	    if (p == NULL) { // we found an empty spot, put dot here
		set_dot(ddbuf, d->x, d->y, d);
		break;
	    }
	}
	d++;
	i++;
    }

    ddwin.ReleaseBuffer();
    ddwin.draw();
}

void
movesand(void)
{
    u_char *ddbuf = ddwin.GetBuffer(&pitch);

    dot *d = dots;
    int i = 0;
    while (i < DOT_CNT) {
	int x = d->x;
	int y = d->y;

	// We first adjust each ones position depending on the directions they
	// are moving. Random number + directions give us something a bit more
	// exciting to look at. Our direction values are between 0 and 3.
	int r = rand() & 0xff;
	if ((r & 3) < d->dir_w) x--;
	r >>= 4;
	if ((r & 3) < d->dir_e) x++;

	r = rand() & 0xff;
	if ((r & 3) < d->dir_n) y--;
	r >>= 4;
	if ((r & 3) < d->dir_s) y++;

	// if we moved off screen, wrap around
	if (x >= SCREEN_WIDTH) x -= SCREEN_WIDTH;
	if (y >= SCREEN_HEIGHT) y -= SCREEN_HEIGHT;
	if (x < 0) x += SCREEN_WIDTH;
	if (y < 0) y += SCREEN_HEIGHT;


	// has our position changed? if so, lets move!
	if (x != d->x || y != d->y) {
	    dot *d2 = get_dot(x, y); // this is the dot we bumped into

	    if (d2 == NULL) { // if we bumped into nothing, just move there
		// remove old dot
		set_dot(ddbuf, d->x, d->y, NULL);

		// move to new position
		d->x = x;
		d->y = y;
		set_dot(ddbuf, x, y, d);

	    } else { // we have bumped into another dot
		// if we hit something that is moving, we change our direction
		// slightly towards the same direction that they are moving

		if (d2->dir_n > d2->dir_s) {
		    // hit something moving up
		    if ((d2->dir_n - d2->dir_s) > 2) { // moving fast?
			if (d->dir_n != 3) d->dir_n++;
			if (d->dir_s != 0) d->dir_s--;
		    }

		    if (d->dir_n != 3) d->dir_n++;
		    if (d->dir_s != 0) d->dir_s--;

		} else if (d2->dir_n < d2->dir_s) {
		    // hit something moving down
		    if ((d2->dir_s - d2->dir_n) > 2) { // moving fast?
			if (d->dir_s != 3) d->dir_s++;
			if (d->dir_n != 0) d->dir_n--;
		    }

		    if (d->dir_s != 3) d->dir_s++;
		    if (d->dir_n != 0) d->dir_n--;

		} else {
		    // We bumped into a dot that isnt moving up or down. This
		    // dot doesnt effect our direction at all.
		}

		// update new color of dot
		// false == don't update position
		set_dot(ddbuf, x, y, d, false);
	    }
	}
	d++;
	i++;
    }

    ddwin.ReleaseBuffer();
    ddwin.draw();
}


LRESULT CALLBACK
vidWndProc2(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);

LRESULT CALLBACK
vidWndProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    switch (uMsg) {
	case WM_CREATE:
	    {
		WNDCLASS     wndclass;
		static char *szAppName2 = "vid2";
		wndclass.style         = 0;
		wndclass.lpfnWndProc   = vidWndProc2;
		wndclass.cbClsExtra    = 0;
		wndclass.cbWndExtra    = 0;
		wndclass.hInstance     = GetModuleHandle(NULL);
		wndclass.hIcon         = LoadIcon (NULL, IDI_APPLICATION);
		wndclass.hCursor       = LoadCursor (NULL, IDC_ARROW);
		wndclass.hbrBackground = (HBRUSH) GetStockObject (WHITE_BRUSH);
		wndclass.lpszMenuName  = NULL;
		wndclass.lpszClassName = szAppName2;
		RegisterClass(&wndclass);
		CreateWindow(szAppName2,		// window class name
			     "",			// window caption
			     WS_CHILD|WS_VISIBLE,
			     0, 0, SCREEN_WIDTH, SCREEN_HEIGHT,
			     hwnd,			// parent window handle
			     NULL,
			     GetModuleHandle(NULL), NULL);
	    }
	    break;

	case WM_DESTROY:
	    PostQuitMessage(0);
	    break;

	case WM_MOVE:
	    ddwin.draw();
	    break;

	case WM_PAINT:
	    {
		PAINTSTRUCT ps;
		BeginPaint(hwnd, &ps);
		EndPaint(hwnd, &ps);
	    }
	    return 0;
    }
    return DefWindowProc(hwnd, uMsg, wParam, lParam);
}

LRESULT CALLBACK
vidWndProc2(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    switch (uMsg) {
	case WM_DESTROY:
	    KillTimer(hwnd, 1);
	    break;

	case WM_CREATE:
	    if (!ddwin.init(hwnd)) {
		MessageBox(NULL, "could not init directdraw\n",
			   "error", MB_OK|MB_ICONERROR);
		PostQuitMessage(0);
	    }
	    initsand();
	    SetTimer(hwnd, 1, 50, NULL);
	    break;

	case WM_TIMER:
	    movesand();
	    break;

	case WM_PAINT:
	    {
		PAINTSTRUCT ps;
		BeginPaint(hwnd, &ps);
		ddwin.draw();
		EndPaint(hwnd, &ps);
	    }
	    return 0;
    }
    return DefWindowProc(hwnd, uMsg, wParam, lParam);
}

int APIENTRY
WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
	LPSTR cmdline, int show)
{
    srand(time(NULL));

    static char *szAppName = "vid";
    WNDCLASS     wndclass;
    wndclass.style         = 0;
    wndclass.lpfnWndProc   = vidWndProc;
    wndclass.cbClsExtra    = 0;
    wndclass.cbWndExtra    = 0;
    wndclass.hInstance     = GetModuleHandle(NULL);
    wndclass.hIcon         = LoadIcon (NULL, IDI_APPLICATION);
    wndclass.hCursor       = LoadCursor (NULL, IDC_ARROW);
    wndclass.hbrBackground = (HBRUSH) GetStockObject (WHITE_BRUSH);
    wndclass.lpszMenuName  = NULL;
    wndclass.lpszClassName = szAppName;
    RegisterClass(&wndclass);
    HWND hwnd = CreateWindow(szAppName,		// window class name
			     "Sand",			// window caption
			     WS_OVERLAPPEDWINDOW|WS_VISIBLE,
			     CW_USEDEFAULT, CW_USEDEFAULT,
			     SCREEN_WIDTH + 8, SCREEN_HEIGHT + 27,
			     NULL,			// parent window handle
			     NULL,
			     GetModuleHandle(NULL), NULL);

    MSG msg;
    while (GetMessage(&msg, NULL, 0, 0)) {
	TranslateMessage(&msg);
	DispatchMessage(&msg);
    }
    return msg.wParam;
}
